import time
import string
from pathlib import Path
import glob
import random

from . import config

SLEEP_TIMEOUT = 0.01

# If a process crashes and doesn't clean its locks, remove them after a timeout
LOCK_TIMEOUT = 40.0



def clean_dead_locks(db_name, ignore=None):
	db_locks = glob.glob(path_str(db_name, "*", "*", "*"))
	for lock in db_locks:
		if lock == ignore:
			continue
		lock_parts = lock.split(".")
		time_ns = lock_parts[-3]
		if time.time_ns() - int(time_ns) > LOCK_TIMEOUT * 1_000_000_000:
			Path(lock).unlink()
			print(f"Found dead lock ({lock}). Remove")



def path_str(db_name, lock_id, time_ns, lock_type):
	path = f"{config.storage_directory}/"
	if "/" in db_name:
		db_name = db_name.split("/")
		db_name[-1] = "." + db_name[-1]
		db_name = "/".join(db_name)
	else:
		path += "."
	return f"{path}{db_name}.{lock_id}.{time_ns}.{lock_type}.lock"



def find_locks(lock_type: str, db_name: str):
	return glob.glob(path_str(db_name, "*", "*", lock_type))



def is_oldest_lock_candidate(lock_id, db_name):
	write_candidates = find_locks("needwrite", db_name)
	write_candidates = [x.split(".")[:-2][-2:] for x in write_candidates]
	oldest_candidate = min(write_candidates, key=lambda x: int(x[1]))[0]
	return oldest_candidate == lock_id



class AbstractLock(object):
	"""
		An abstract lock doesn't do anything by itself. A subclass of it needs to
		call super().__init__(...) and then only exit when the lock is aquired.
	"""
	def __init__(self, db_name):
		"""
			If key is None, create a random identifier.
			if time_ns is None, initialize it to the current time.
		"""
		# Create a random id (62^5 = 916.132.832 possibilities)
		self.id = "".join(random.choices(string.ascii_letters + string.digits, k=5))
		self.time_ns = time.time_ns()
		self.db_name = db_name
		self.path = None

	def unlock(self):
		self.path.unlink()
		self.path = None



class ReadLock(AbstractLock):
	def __init__(self, db_name):
		super().__init__(db_name)
		has_read_path_str = path_str(db_name, self.id, self.time_ns, "hasread")
		self.path = Path(has_read_path_str)
		while True:
			clean_dead_locks(db_name, ignore=has_read_path_str)
			if len(find_locks("*write", db_name)) == 0:
				self.path.touch()
				return
			time.sleep(SLEEP_TIMEOUT)



class WriteLock(AbstractLock):
	def __init__(self, db_name):
		super().__init__(db_name)
		need_write_path_str = path_str(db_name, self.id, self.time_ns, "needwrite")
		need_write_path = Path(need_write_path_str)
		need_write_path.touch()
		self.path = Path(path_str(db_name, self.id, self.time_ns, "haswrite"))
		while True:
			clean_dead_locks(db_name, ignore=need_write_path_str)
			if is_oldest_lock_candidate(self.id, db_name) and len(find_locks("has*", db_name)) == 0:
				self.path.touch()
				need_write_path.unlink()
				return
			time.sleep(SLEEP_TIMEOUT)


from __future__ import annotations
import glob
from . import utils, config
from .locking import find_locks

from path_dict import PathDict



def get_db_names(pattern: str):
	dbs_ddb = glob.glob(f"{config.storage_directory}/{pattern}.ddb")
	dbs_json = glob.glob(f"{config.storage_directory}/{pattern}.json")
	dbs_all = dbs_ddb + dbs_json
	dbs_all = [d.replace(config.storage_directory + "/", "") for d in dbs_all]
	dbs_all = [d.replace(".ddb", "") for d in dbs_all]
	dbs_all = [d.replace(".json", "") for d in dbs_all]
	return dbs_all



class DDBSession(object):
	"""
		Enter: "with DDBSession(db_name) as session, dict:
		Then modify dict as required.
		Save: session.save_changes()
		Discard: session.discard_changes():
		One of both need to be called
	"""
	def __init__(self, db_name: str, as_PathDict: bool = False):
		self.db_name = db_name
		self.as_PathDict = as_PathDict
		self.in_session = False


	def __enter__(self):
		"""
			Any number of read tasks can be carried out in parallel.
			Each read task creates a read lock while reading, to signal that it is reading.

			As soon as a session starts, it writes a wants-to-write lock,
			No new read tasks will be allowed. When all read tasks are done, the session aquire the write lock.
			Now, it can savely read and write while all other tasks wait.
		"""
		if len(find_locks("haswrite", self.db_name)):
			raise Exception("Never access the same db again during an open session!")
		self.write_lock = utils.WriteLock(self.db_name)
		self.in_session = True
		try:
			self.dict = utils.unprotected_read_json_as_dict(self.db_name)
			if self.as_PathDict:
				self.dict = PathDict(self.dict)
		except BaseException:
			self.write_lock.unlock()
			raise
		return self, self.dict


	def __exit__(self, type, value, tb):
		self.write_lock.unlock()
		self.write_lock = None
		self.in_session = False


	def write(self):
		if not self.in_session:
			raise Exception("Only call save_changes() inside a with statement.")
		if self.as_PathDict:
			utils.unprotected_write_dict_as_json(self.db_name, self.dict.data)
		else:
			utils.unprotected_write_dict_as_json(self.db_name, self.dict)



class DDBMultiSession(object):
	def __init__(self, pattern: str, as_PathDict: bool = False):
		self.db_names = get_db_names(pattern)
		self.as_PathDict = as_PathDict
		self.in_session = False


	def __enter__(self):
		for db_name in self.db_names:
			if len(find_locks("haswrite", db_name)):
				raise Exception("Never access the same db again during an open session!")

		self.write_locks = [utils.WriteLock(x) for x in self.db_names]
		self.in_session = True
		try:
			self.dicts = {n: utils.unprotected_read_json_as_dict(n) for n in self.db_names}
			if self.as_PathDict:
				self.dicts = PathDict(self.dicts)
		except BaseException:
			for write_lock in self.write_locks:
				write_lock.unlock()
			raise
		return self, self.dicts


	def __exit__(self, type, value, tb):
		for write_lock in self.write_locks:
			write_lock.unlock()
		self.write_lock = None
		self.in_session = False


	def write(self):
		if not self.in_session:
			raise Exception("Only call save_changes() inside a with statement.")
		for db_name in self.db_names:
			if self.as_PathDict:
				utils.unprotected_write_dict_as_json(db_name, self.dicts[db_name].data)
			else:
				utils.unprotected_write_dict_as_json(db_name, self.dicts[db_name])



def _to_path_if_tuple(s):
	if isinstance(s, tuple) and len(s) > 0:
		return "/".join(list(s))
	return s



class SubModel(PathDict):
	def __init__(self, key: str, initial_value=None):
		"""
			Initialize with the initial_value dict or PathDict if it is given,
			otherwise initi
		"""
		if key is None:
			raise Exception("key can not be None")

		self.key = key
		self.db_name = _to_path_if_tuple(self.db_name)
		if initial_value is None:
			self.read()
		else:
			if not isinstance(initial_value, dict):
				if not isinstance(initial_value, PathDict):
					raise Exception("initial_value must be a dict or PathDict")
			super().__init__(initial_value)


	def session(self):
		return DDBSession(self.db_name, as_PathDict=True)


	def read(self):
		"""
			Load the contents of the object from the database
		"""
		self.file_db = utils.protected_read_json_as_dict(self.db_name)
		if self.file_db is None or self.key not in self.file_db:
			return None
		self.data = self.file_db.get(self.key, None)
		return self


import glob
from DictDataBase.objects import DDBSession, DDBMultiSession, PathDict
from DictDataBase import utils
from . import config
from .objects import SubModel
from .locking import find_locks

from path_dict import PathDict



def expand_find_path_pattern(pattern):
	"""
		:param str pattern: The pattern to expand.
		Fot a tuple of path items, expand it to a list of all real paths.
		An item can be some string, a wildcard "*" or a list to select specific paths.
	"""
	res = [[]]
	for item in pattern:
		if isinstance(item, str):
			res = [r + [item] for r in res]
		if isinstance(item, list):
			res = [r + [list_item] for list_item in item for r in res]
	return [f for r in res for f in find("/".join(r))]



def _to_path_if_tuple(s):
	if isinstance(s, tuple) and len(s) > 0:
		return "/".join(list(s))
	return s



def find(*pattern):
	"""
		Find multiple dbs with a glob pattern,
		and return their names as a list.
	"""
	pattern = _to_path_if_tuple(pattern)
	dbs_ddb = glob.glob(f"{config.storage_directory}/{pattern}.ddb")
	dbs_json = glob.glob(f"{config.storage_directory}/{pattern}.json")
	dbs_all = dbs_ddb + dbs_json
	for trim in [f"{config.storage_directory}/", ".ddb", ".json"]:
		dbs_all = [d.replace(trim, "") for d in dbs_all]
	return dbs_all



def exists(*pattern):
	"""
		Tells if a glob pattern finds a db.
	"""
	pattern = _to_path_if_tuple(pattern)
	return len(find(pattern)) > 0



def read(*name, as_PathDict: bool = False):
	name = _to_path_if_tuple(name)

	if len(find_locks("haswrite", name)) > 0:
		raise Exception("Never access the same db again during an open session!")

	db = utils.protected_read_json_as_dict(name)
	if as_PathDict:
		return PathDict(db)
	return db


def multiread(*pattern, as_PathDict: bool = False):
	"""
		Mutliread reads multiple dbs and returns them as a single dict or PathDict.
		Path components can be "*" (all), a specific name or a list (only those from list).
	"""
	pattern_paths = expand_find_path_pattern(pattern)
	res = {}
	for db_name in pattern_paths:
		res[db_name] = utils.protected_read_json_as_dict(db_name)
	if as_PathDict:
		return PathDict(res)
	return res



def create(*db_name, db={}):
	db_name = _to_path_if_tuple(db_name)
	if isinstance(db, PathDict):
		utils.protected_write_dict_as_json(db_name, db.dict)
	else:
		utils.protected_write_dict_as_json(db_name, db)



def delete(*db_name):
	db_name = _to_path_if_tuple(db_name)
	utils.protected_delete(db_name)


def session(*db_name, as_PathDict: bool = False):
	db_name = _to_path_if_tuple(db_name)
	return DDBSession(db_name, as_PathDict=as_PathDict)

def multisession(*pattern, as_PathDict: bool = False):
	"""
		Open multiple files at once using a glob pattern, like "user*".
		Mutliple arguments are allowed to access folders,
		so multisession(f"users/{user_id}") is equivalent
		to multisession("users", user_id).
	"""
	pattern = _to_path_if_tuple(pattern)
	return DDBMultiSession(pattern, as_PathDict=as_PathDict)




class Model(PathDict):
	"""
		Used to collect functions of a particular db file or
		set of files
	"""
	def __init__(self, *db_name):
		if not exists(*db_name):
			return None
		db_name = _to_path_if_tuple(db_name)
		self.db_name = db_name
		self.reread()


	def reread(self):
		self.data = utils.protected_read_json_as_dict(self.db_name)

	def write(self):
		utils.protected_write_dict_as_json(self.db_name, self.data)


import os
import json
import zlib

from .locking import ReadLock, WriteLock
from . import config



def db_paths(db_name: str) -> (str, bool, str, bool):
	base = f"{config.storage_directory}/{db_name}"
	j, d = f"{base}.json", f"{base}.ddb"
	return j, os.path.exists(j), d, os.path.exists(d)



def unprotected_read_json_as_dict(db_name: str) -> dict:
	"""
		Read the file at db_path from the configured storage directory.
		Make sure the file exists!
	"""
	json_path, json_exists, ddb_path, ddb_exists = db_paths(db_name)


	read_mode, db_path = None, None
	if json_exists and ddb_exists:
		raise Exception(f"DB Inconsistency: \"{db_name}\" exists as .json and .ddb")
	elif json_exists:
		read_mode, db_path = "r", json_path
	elif ddb_exists:
		read_mode, db_path = "rb", ddb_path
	else:
		raise Exception(f"DB \"{db_name}\" does not exist.")

	with open(db_path, read_mode) as f:
		f_str = f.read()
		if read_mode == "rb":
			f_str = zlib.decompress(f_str).decode()
		return json.loads(f_str)



def protected_read_json_as_dict(db_name: str):
	"""
		Ensure that reading only starts when there is no writing,
		and that while reading, no writing will happen.
		Otherwise, wait.
	"""

	_, json_exists, _, ddb_exists = db_paths(db_name)
	if not json_exists and not ddb_exists:
		return None
	# Wait in any write lock case, "need" or "has".
	lock = ReadLock(db_name)
	res = unprotected_read_json_as_dict(db_name)
	lock.unlock()
	return res





def unprotected_write_dict_as_json(db_name: str, db: dict):
	"""
		Write the dict db dumped as a json string
		to the file of the db_path.
	"""
	db_dump = None
	if config.pretty_json_files and not config.use_compression:
		db_dump = json.dumps(db, indent="\t", sort_keys=True)
	else:
		db_dump = json.dumps(db)

	json_path, json_exists, ddb_path, ddb_exists = db_paths(db_name)

	write_mode, db_path = None, None
	if config.use_compression:
		write_mode, db_path = "wb", ddb_path
		if json_exists:
			os.remove(json_path)
	else:
		write_mode, db_path = "w+", json_path
		if ddb_exists:
			os.remove(ddb_path)

	with open(db_path, write_mode) as f:
		if config.use_compression:
			db_dump = zlib.compress(db_dump.encode(), 1)
		f.write(db_dump)



def protected_write_dict_as_json(db_name: str, db: dict):
	"""
		Ensures that writing only starts if there is no reading or writing in progress.
	"""
	dirname = os.path.dirname(f"{config.storage_directory}/{db_name}.any")
	os.makedirs(dirname, exist_ok=True)

	write_lock = WriteLock(db_name)
	unprotected_write_dict_as_json(db_name, db)
	write_lock.unlock()


def protected_delete(db_name: str):
	"""
		Ensures that deleting only starts if there is no reading or writing in progress.
	"""
	json_path, json_exists, ddb_path, ddb_exists = db_paths(db_name)
	if not json_exists and not ddb_exists:
		return None
	write_lock = WriteLock(db_name)
	if json_exists:
		os.remove(json_path)
	if ddb_exists:
		os.remove(ddb_path)
	write_lock.unlock()